@@ -2,6 +2,9 @@
 
 #include <stdio.h>
 #include <string.h>
+#include <sys/types.h>
+#include <sys/stat.h>
+#include <unistd.h>
 
 #include "structmember.h"
 #include "otama.h"
@@ -21,6 +24,7 @@
 
 static PyObject *PyExc_OtamaError;
 static PyTypeObject OtamaObjectType;
+static PyTypeObject OtamaFeatureRawObjectType;
 static void pyobj2variant(PyObject *object, otama_variant_t *var);
 
 /* Otama Object */
@@ -29,6 +33,11 @@ typedef struct {
     otama_t *otama;
 } OtamaObject;
 
+typedef struct {
+    OtamaObject base;
+    otama_feature_raw_t *raw;
+} OtamaFeatureRawObject;
+
 
 static PyObject *
 variant2pyobj(otama_variant_t *var)
@@ -38,10 +47,10 @@ variant2pyobj(otama_variant_t *var)
             return PyLong_FromLong(otama_variant_to_int(var));
         case OTAMA_VARIANT_TYPE_FLOAT:
             return PyFloat_FromDouble(otama_variant_to_float(var));
-	    case OTAMA_VARIANT_TYPE_STRING:
+        case OTAMA_VARIANT_TYPE_STRING:
             return PyString_FromString(otama_variant_to_string(var));
-	    case OTAMA_VARIANT_TYPE_ARRAY: {
-		    long count = otama_variant_array_count(var);
+        case OTAMA_VARIANT_TYPE_ARRAY: {
+            long count = otama_variant_array_count(var);
             int i;
             PyObject *tuple = PyTuple_New(count);
             for (i = 0; i < count; ++i) {
@@ -67,7 +76,7 @@ variant2pyobj(otama_variant_t *var)
         }
         case OTAMA_VARIANT_TYPE_NULL:
         default:
-            printf("not implementation");
+            printf("not implementation\n");
             break;
     }
 
@@ -101,10 +110,20 @@ pyobj2variant(PyObject *object, otama_variant_t *var)
         otama_variant_set_int(var, PyLong_AsLong(object));
     }
     else if (PyString_Check(object)) {
-        otama_variant_set_string(var, PyString_AsString(object));
+        if (strlen(PyString_AsString(object)) == PyString_GET_SIZE(object)) {
+            otama_variant_set_string(var, PyString_AsString(object));
+        }
+        else {
+            otama_variant_set_binary(var, PyString_AsString(object), PyString_GET_SIZE(object));
+        }
     }
     else if (PyTuple_Check(object)) {
-        printf("tuple\n");
+        int len = PyTuple_Size(object), i;
+        otama_variant_set_array(var);
+        for (i = 0; i < len; ++i) {
+            PyObject *elm = PyTuple_GetItem(object, i);
+            pyobj2variant(elm, otama_variant_array_at(var, i));
+        }
     }
     else if (PyDict_Check(object)) {
         otama_variant_set_hash(var);
@@ -120,7 +139,12 @@ pyobj2variant(PyObject *object, otama_variant_t *var)
         }
     }
     else {
-        printf("%s not support\n", __FUNCTION__);
+        if (PyObject_IsInstance(object, (PyObject *)&OtamaFeatureRawObjectType)) {
+            otama_variant_set_pointer(var, ((OtamaFeatureRawObject *)object)->raw);
+        }
+        else {
+            otama_variant_set_null(var);
+        }
     }
 }
 
@@ -205,7 +229,7 @@ OtamaObject_new(PyTypeObject *type, PyObject *args)
                 var = otama_variant_new(pool);
 
                 pyobj2variant(config, var);
-                otama_open_opt(&self->otama, var);
+                ret = otama_open_opt(&self->otama, var);
 
                 otama_variant_pool_free(&pool);
             }
@@ -278,7 +302,7 @@ static PyObject *
 OtamaObject_close(OtamaObject *self)
 {
     if (self->otama) {
-        otama_close(&(self->otama));
+        otama_close(&self->otama);
         self->otama = NULL;
     }
 
@@ -348,6 +372,14 @@ OtamaObject_search(OtamaObject *self, PyObject *args)
 
     if (PyString_Check(data)) {
         const char *_tmp = PyString_AsString(data);
+        char _err_tmp[120] = "not exist file ";
+        struct stat st;
+        if (stat(_tmp, &st)) {
+            otama_variant_pool_free(&pool);
+            strcat(_err_tmp, _tmp);
+            PyErr_SetString(PyExc_IOError, _err_tmp);
+            return NULL;
+        }
         ret = otama_search_file(self->otama, &results, num, _tmp);
     }
     else {
@@ -368,6 +400,43 @@ OtamaObject_search(OtamaObject *self, PyObject *args)
     return result_tuple;
 }
 
+static PyObject *
+OtamaObject_similarity(OtamaObject *self, PyObject *args)
+{
+    PyObject *data1, *data2;
+    otama_status_t ret;
+    otama_variant_pool_t *pool;
+    otama_variant_t *var1, *var2;
+    float similarity = 0.0f;
+
+    if (!PyArg_ParseTuple(args, "OO", &data1, &data2)) {
+        PyErr_SetString(PyExc_TypeError, "argument error");
+        return NULL;
+    }
+
+    if (!(PyDict_Check(data1) && PyDict_Check(data2))) {
+        PyErr_SetString(PyExc_OtamaError, "invalid argument type");
+        return NULL;
+    }
+
+    pool = otama_variant_pool_alloc();
+    var1 = otama_variant_new(pool);
+    var2 = otama_variant_new(pool);
+
+    pyobj2variant(data1, var1);
+    pyobj2variant(data2, var2);
+
+	ret = otama_similarity(self->otama, &similarity, var1, var2);
+	if (ret != OTAMA_STATUS_OK) {
+		otama_variant_pool_free(&pool);
+		otamapy_raise(ret);
+		return NULL;
+	}
+	otama_variant_pool_free(&pool);
+
+	return PyFloat_FromDouble(similarity);
+}
+
 static PyObject *
 OtamaObject_insert(OtamaObject *self, PyObject *args)
 {
@@ -437,6 +506,81 @@ OtamaObject_remove(OtamaObject *self, PyObject *args)
     Py_RETURN_NONE;
 }
 
+static PyObject *
+OtamaObject_exists(OtamaObject *self, PyObject *args)
+{
+    PyObject *id;
+    otama_status_t ret;
+    otama_id_t otama_id;
+    int result = 0;
+
+    if (!PyArg_ParseTuple(args, "O", &id)) {
+        PyErr_SetString(PyExc_TypeError, "argument error");
+        return NULL;
+    }
+
+    ret = otama_id_hexstr2bin(&otama_id, PyString_AsString(id));
+    if (ret != OTAMA_STATUS_OK) {
+        otamapy_raise(ret);
+        return NULL;
+    }
+
+    ret = otama_exists(self->otama, &result, &otama_id);
+    if (ret != OTAMA_STATUS_OK) {
+        otamapy_raise(ret);
+        return NULL;
+    }
+
+    return result == 0 ? PyBool_FromLong(0) : PyBool_FromLong(1);
+}
+
+static PyObject *
+OtamaObject_feature_raw(OtamaObject *self, PyObject *args)
+{
+    otama_status_t ret;
+    otama_feature_raw_t *raw;
+    PyObject *pyraw, *query;
+    otama_variant_pool_t *pool;
+    otama_variant_t *var;
+
+    if (!PyArg_ParseTuple(args, "O", &query)) {
+        PyErr_SetString(PyExc_TypeError, "argument error");
+        return NULL;
+    }
+
+    if (!PyDict_Check(query)) {
+        PyErr_SetString(PyExc_TypeError, "invalid argument");
+        return NULL;
+    }
+
+    pool = otama_variant_pool_alloc();
+    var = otama_variant_new(pool);
+
+    pyobj2variant(query, var);
+
+    ret = otama_feature_raw(self->otama, &raw, var);
+    if (ret != OTAMA_STATUS_OK) {
+        otama_variant_pool_free(&pool);
+        otamapy_raise(ret);
+        return NULL;
+    }
+    otama_variant_pool_free(&pool);
+
+    pyraw = PyType_GenericNew(&OtamaFeatureRawObjectType, NULL, NULL);
+    ((OtamaFeatureRawObject *)pyraw)->raw = raw;
+
+    return pyraw;
+}
+
+static PyObject *
+OtamaFeatureRawObject_dispose(OtamaFeatureRawObject *self)
+{
+    otama_feature_raw_free(&self->raw);
+    self->raw = NULL;
+
+    Py_RETURN_NONE;
+}
+
 static PyMethodDef OtamaObject_methods[] = {
     {"open", (PyCFunction)OtamaObject_open, METH_VARARGS|METH_STATIC,
      "open Otama"},
@@ -454,6 +598,12 @@ static PyMethodDef OtamaObject_methods[] = {
      "remove id from Otama DataBase"},
     {"search", (PyCFunction)OtamaObject_search, METH_VARARGS,
      "search from Otama DataBase"},
+    {"similarity", (PyCFunction)OtamaObject_similarity, METH_VARARGS,
+     "check similarity"},
+    {"exists", (PyCFunction)OtamaObject_exists, METH_VARARGS,
+     "exist image in Otama DataBase"},
+    {"feature_raw", (PyCFunction)OtamaObject_feature_raw, METH_VARARGS,
+     "return feature raw value"},
     {NULL, NULL, 0, NULL}
 };
 
@@ -508,6 +658,61 @@ static PyTypeObject OtamaObjectType = {
     (newfunc)OtamaObject_new,                   /* tp_new */
 };
 
+static PyMethodDef OtamaFeatureRawObject_methods[] = {
+    {"dispose", (PyCFunction)OtamaFeatureRawObject_dispose, METH_NOARGS,
+     "free resource"},
+    {NULL, NULL, 0, NULL}
+};
+
+static PyMemberDef OtamaFeatureRawObject_members[] = {
+    {NULL}
+};
+
+static PyTypeObject OtamaFeatureRawObjectType = {
+#ifdef PY3
+    PyVarObject_HEAD_INIT(NULL, 0)
+#else
+    PyObject_HEAD_INIT(NULL)
+    0,                                          /* ob_size */
+#endif
+    "otama.OtamaFeatureRaw",                    /* tp_name */
+    sizeof(OtamaFeatureRawObject),              /* tp_basicsize */
+    0,
+    (destructor)Otama_dealloc,                  /* tp_dealloc */
+    0,
+    0,
+    0,
+    0,
+    0,
+    0,
+    0,
+    0,
+    0,
+    0,
+    0,
+    0,
+    0,
+    0,
+    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_BASETYPE,   /* tp_flags */
+    "OtamaFeatureRaw objects",                  /* tp_doc */
+    0,
+    0,
+    0,
+    0,
+    0,
+    0,
+    OtamaFeatureRawObject_methods,              /* tp_methods */
+    OtamaFeatureRawObject_members,              /* tp_members */
+    0,
+    0,
+    0,
+    0,
+    0,
+    0,
+    (initproc)OtamaObject_init,                 /* tp_init */
+    0,                                          /* tp_alloc */
+    (newfunc)OtamaObject_new,                   /* tp_new */
+};
 
 static PyMethodDef OtamaMethods[] = {
     {NULL, NULL, 0, NULL}
@@ -551,11 +756,17 @@ initotama(void)
     if (PyType_Ready(&OtamaObjectType) < 0)
         OTAMAPY_INIT_ERROR;
 
-    PyExc_OtamaError = PyErr_NewException("otama.error", NULL, NULL);
+    if (PyType_Ready(&OtamaFeatureRawObjectType) < 0)
+        OTAMAPY_INIT_ERROR;
+
+    PyExc_OtamaError = PyErr_NewException("otama.OtamaError", NULL, NULL);
 
     Py_INCREF(&OtamaObjectType);
     PyModule_AddObject(module, "Otama", (PyObject *)&OtamaObjectType);
 
+    Py_INCREF(&OtamaFeatureRawObjectType);
+    PyModule_AddObject(module, "OtamaFeatureRaw", (PyObject *)&OtamaFeatureRawObjectType);
+
     Py_INCREF(PyExc_OtamaError);
     PyModule_AddObject(module, "error", (PyObject *)&OtamaObjectType);
 
